/*
 * Copyright IBM Corp. 2001, 2018
 *
 * s390-tools is free software; you can redistribute it and/or modify
 * it under the terms of the MIT license. See LICENSE for details.
 */

#ifndef STDOUT_H
#define STDOUT_H

int stdout_write_dump(void);

#endif /* STDOUT_H */
